@@ -2,6 +2,7 @@
 
 #include <algorithm>
 #include <cstdlib>
+#include <limits>
 
 #include <string.h>
 
@@ -132,7 +133,12 @@ void buffer::append(unsigned char const* data, size_t len)
 	// until after appending in case of append from own memory
 	unsigned char* old{};
 	if (capacity_ - (pos_ - data_) - size_ < len) {
-		if (capacity_ - size_ > len) {
+		if (capacity_ - size_ >= len) {
+			// Also offset data in case of self-assignment
+			if (data >= pos_ && data < (pos_ + size_)) {
+				data -= pos_ - data_;
+			}
+
 			memmove(data_, pos_, size_);
 			pos_ = data_;
 		}
@@ -152,8 +158,10 @@ void buffer::append(unsigned char const* data, size_t len)
 		}
 	}
 
-	memcpy(pos_ + size_, data, len);
-	size_ += len;
+	if (len) {
+		memcpy(pos_ + size_, data, len);
+		size_ += len;
+	}
 
 	delete [] old;
 }
@@ -218,4 +226,13 @@ bool buffer::operator==(buffer const& rhs) const
 	return memcmp(get(), rhs.get(), size()) == 0;
 }
 
+std::string_view buffer::to_view() const
+{
+	if (!size()) {
+		return {};
+	}
+
+	return {reinterpret_cast<char const*>(get()), size()};
+}
+
 }
