@@ -1,7 +1,9 @@
 #include "libfilezilla/libfilezilla.hpp"
 #include "libfilezilla/file.hpp"
 
-#ifndef FZ_WINDOWS
+#ifdef FZ_WINDOWS
+#include "windows/security_descriptor_builder.hpp"
+#else
 #include <errno.h>
 #include <sys/stat.h>
 #include <sys/types.h>
@@ -16,19 +18,40 @@ file::file(native_string const& f, mode m, creation_flags d)
 	open(f, m, d);
 }
 
+file::file(file::file_t fd)
+	: fd_(fd)
+{
+}
+
 file::~file()
 {
 	close();
 }
 
 #ifdef FZ_WINDOWS
+file::file(file && op) noexcept
+	: fd_{op.fd_}
+{
+	op.fd_ = INVALID_HANDLE_VALUE;
+}
+
+file& file::operator=(file && op) noexcept
+{
+	if (this != &op) {
+		close();
+		fd_ = op.fd_;
+		op.fd_ = INVALID_HANDLE_VALUE;
+	}
+	return *this;
+}
+
 bool file::open(native_string const& f, mode m, creation_flags d)
 {
 	close();
 
 	DWORD dispositionFlags;
 	if (m == writing) {
-		if (d == empty) {
+		if (d & empty) {
 			dispositionFlags = CREATE_ALWAYS;
 		}
 		else {
@@ -44,70 +67,49 @@ bool file::open(native_string const& f, mode m, creation_flags d)
 		shareMode |= FILE_SHARE_WRITE;
 	}
 
-	if (d & current_user_only) {
-		HANDLE token{INVALID_HANDLE_VALUE};
-		bool res = OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, &token);
-		if (!res) {
+	SECURITY_ATTRIBUTES attr{};
+	attr.bInheritHandle = false;
+	attr.nLength = sizeof(SECURITY_ATTRIBUTES);
+
+	security_descriptor_builder sdb;
+	if (d & (current_user_only | current_user_and_admins_only)) {
+		sdb.add(security_descriptor_builder::self);
+		if ((d & current_user_and_admins_only) == current_user_and_admins_only) {
+			sdb.add(security_descriptor_builder::administrators);
+		}
+
+		auto sd = sdb.get_sd();
+		if (!sd) {
 			return false;
 		}
-
-		DWORD needed{};
-		GetTokenInformation(token, TokenUser, NULL, 0, &needed);
-		if (GetLastError() != ERROR_INSUFFICIENT_BUFFER) {
-			return false;
-		}
-
-		PTOKEN_USER tu = static_cast<PTOKEN_USER>(malloc(needed));
-		if (tu) {
-			if (GetTokenInformation(token, TokenUser, tu, needed, &needed)) {
-				needed = sizeof(ACL) + ((sizeof(ACCESS_ALLOWED_ACE) - sizeof(DWORD)) + GetLengthSid(tu->User.Sid));
-				PACL acl = static_cast<PACL>(malloc(needed));
-				if (acl) {
-					if (InitializeAcl(acl, needed, ACL_REVISION)) {
-						if (AddAccessAllowedAce(acl, ACL_REVISION, GENERIC_ALL | STANDARD_RIGHTS_ALL | SPECIFIC_RIGHTS_ALL, tu->User.Sid)) {
-							SECURITY_DESCRIPTOR sd;
-							InitializeSecurityDescriptor(&sd, SECURITY_DESCRIPTOR_REVISION);
-							SetSecurityDescriptorDacl(&sd, TRUE, acl, FALSE);
-							SetSecurityDescriptorOwner(&sd, tu->User.Sid, FALSE);
-							SetSecurityDescriptorGroup(&sd, NULL, FALSE);
-							SetSecurityDescriptorSacl(&sd, FALSE, NULL, FALSE);
-
-							SECURITY_ATTRIBUTES attr;
-							attr.bInheritHandle = false;
-							attr.nLength = sizeof(SECURITY_ATTRIBUTES);
-							attr.lpSecurityDescriptor = &sd;
-
-							hFile_ = CreateFile(f.c_str(), (m == reading) ? GENERIC_READ : GENERIC_WRITE, shareMode, &attr, dispositionFlags, FILE_FLAG_SEQUENTIAL_SCAN, nullptr);
-						}
-					}
-					free(acl);
-				}
-			}
-			free(tu);
-		}
-		CloseHandle(token);
+		attr.lpSecurityDescriptor = sd;
 	}
-	else {
-		hFile_ = CreateFile(f.c_str(), (m == reading) ? GENERIC_READ : GENERIC_WRITE, shareMode, nullptr, dispositionFlags, FILE_FLAG_SEQUENTIAL_SCAN, nullptr);
-	}
-	
-	return hFile_ != INVALID_HANDLE_VALUE;
+	fd_ = CreateFile(f.c_str(), (m == reading) ? GENERIC_READ : GENERIC_WRITE, shareMode, &attr, dispositionFlags, FILE_FLAG_SEQUENTIAL_SCAN, nullptr);
+
+	return fd_ != INVALID_HANDLE_VALUE;
 }
 
 void file::close()
 {
-	if (hFile_ != INVALID_HANDLE_VALUE) {
-		CloseHandle(hFile_);
-		hFile_ = INVALID_HANDLE_VALUE;
+	if (fd_ != INVALID_HANDLE_VALUE) {
+		CloseHandle(fd_);
+		fd_ = INVALID_HANDLE_VALUE;
 	}
 }
 
+file::file_t file::detach()
+{
+	file_t fd = fd_;
+	fd_ = INVALID_HANDLE_VALUE;
+	return fd;
+}
+
 int64_t file::size() const
 {
 	int64_t ret = -1;
 
 	LARGE_INTEGER size{};
-	if (GetFileSizeEx(hFile_, &size)) {
+	if (GetFileSizeEx(fd_, &size)) {
 		ret = static_cast<int64_t>(size.QuadPart);
 	}
 	return ret;
@@ -129,7 +131,7 @@ int64_t file::seek(int64_t offset, seek_mode m)
 	}
 
 	LARGE_INTEGER newPos{};
-	if (SetFilePointerEx(hFile_, dist, &newPos, method)) {
+	if (SetFilePointerEx(fd_, dist, &newPos, method)) {
 		ret = newPos.QuadPart;
 	}
 	return ret;
@@ -137,7 +139,7 @@ int64_t file::seek(int64_t offset, seek_mode m)
 
 bool file::truncate()
 {
-	return !!SetEndOfFile(hFile_);
+	return !!SetEndOfFile(fd_);
 }
 
 int64_t file::read(void *buf, int64_t count)
@@ -145,7 +147,7 @@ int64_t file::read(void *buf, int64_t count)
 	int64_t ret = -1;
 
 	DWORD read = 0;
-	if (ReadFile(hFile_, buf, static_cast<DWORD>(count), &read, nullptr)) {
+	if (ReadFile(fd_, buf, static_cast<DWORD>(count), &read, nullptr)) {
 		ret = static_cast<int64_t>(read);
 	}
 
@@ -157,7 +159,7 @@ int64_t file::write(void const* buf, int64_t count)
 	int64_t ret = -1;
 
 	DWORD written = 0;
-	if (WriteFile(hFile_, buf, static_cast<DWORD>(count), &written, nullptr)) {
+	if (WriteFile(fd_, buf, static_cast<DWORD>(count), &written, nullptr)) {
 		ret = static_cast<int64_t>(written);
 	}
 
@@ -166,7 +168,7 @@ int64_t file::write(void const* buf, int64_t count)
 
 bool file::opened() const
 {
-	return hFile_ != INVALID_HANDLE_VALUE;
+	return fd_ != INVALID_HANDLE_VALUE;
 }
 
 bool remove_file(native_string const& name)
@@ -181,11 +183,27 @@ bool remove_file(native_string const& name)
 
 bool file::fsync()
 {
-	return FlushFileBuffers(hFile_) != 0;
+	return FlushFileBuffers(fd_) != 0;
 }
 
 #else
 
+file::file(file && op) noexcept
+	: fd_{op.fd_}
+{
+	op.fd_ = -1;
+}
+
+file& file::operator=(file && op) noexcept
+{
+	if (this != &op) {
+		close();
+		fd_ = op.fd_;
+		op.fd_ = -1;
+	}
+	return *this;
+}
+
 bool file::open(native_string const& f, mode m, creation_flags d)
 {
 	close();
@@ -196,12 +214,12 @@ bool file::open(native_string const& f, mode m, creation_flags d)
 	}
 	else {
 		flags |= O_WRONLY | O_CREAT;
-		if (d == empty) {
+		if (d & empty) {
 			flags |= O_TRUNC;
 		}
 	}
 	int mode = S_IRUSR | S_IWUSR;
-	if (!(d & current_user_only)) {
+	if (!(d & (current_user_only | current_user_and_admins_only))) {
 		mode |= S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH;
 	}
 	fd_ = ::open(f.c_str(), flags, mode);
@@ -223,6 +241,13 @@ void file::close()
 	}
 }
 
+file::file_t file::detach()
+{
+	file_t fd = fd_;
+	fd_ = -1;
+	return fd;
+}
+
 int64_t file::size() const
 {
 	int64_t ret = -1;
