@@ -5,7 +5,7 @@
 #include "time.hpp"
 
 #ifdef FZ_WINDOWS
-#include "private/windows.hpp"
+#include "glue/windows.hpp"
 #else
 #include <dirent.h>
 #endif
@@ -103,7 +103,12 @@ public:
 	/// \brief Begins enumerating a directory.
 	///
 	/// \param dirs_only If true, only directories are enumerated.
-	result begin_find_files(native_string path, bool dirs_only = false);
+	result begin_find_files(native_string path, bool dirs_only = false, bool query_symlink_targets = true);
+
+#ifdef FZ_UNIX
+	// Takes ownership of fd
+	result begin_find_files(int fd, bool dirs_only = false, bool query_symlink_targets = true);
+#endif
 
 	/// Gets the next file in the directory. Call until it returns false.
 	bool get_next_file(native_string& name);
@@ -141,6 +146,20 @@ private:
 
 	// State for directory enumeration
 	bool dirs_only_{};
+	bool query_symlink_targets_{true};
+};
+
+enum class mkdir_permissions
+{
+	/// Normal permissions, on MSW this means inheriting the parent's permissions,
+	/// on *nix the current umask is applied.
+	normal, 
+
+	/// Only current user
+	cur_user, 
+
+	/// Only current user and administrators
+	cur_user_and_admins 
 };
 
 /**
@@ -155,13 +174,28 @@ private:
  * \param absolute_path The directory to create
  * \param recurse If true, creation is recursive. If false, only the last segment gets
  *                created, all other components already need to exist.
- * \param current_user_only If true when creating the last component, its permissions will be
- *                          restricted to the current user. Otherwise default permissions are used.
- *                          Default permissions on  MSW this means inheriting the parent's permissions,
- *                          on *nix the current umask is applied.
- * \param last_created      If non-null, receives the longest sub-path that was created
+ * \param permissions When creating the last component, these permissions are applied.
+ *                    Does not change existing permissions or permissings of newly
+ *                    created parent directories during recursive mkdir.
+ * \param last_created If non-null, receives the longest sub-path that was created
  */
-result FZ_PUBLIC_SYMBOL mkdir(native_string const& absolute_path, bool recurse, bool current_user_only = false, native_string * last_created = nullptr);
+result FZ_PUBLIC_SYMBOL mkdir(native_string const& absolute_path, bool recurse, mkdir_permissions permissions = mkdir_permissions::normal, native_string * last_created = nullptr);
+
+/**
+ * \brief Rename/move the passed file or directory
+ *
+ * Can rename both files and directories within same filesystem. Can
+ * optionally rename files across filesystems
+ *
+ * If renaming a file, the target file is overwritten.
+ *
+ * Directories cannot be renamed/moved across differnet filesystems/mount points
+ *
+ * param allow_copy If true, files, but not directories, can be moved across
+ *                  file system boundaries. It first copies the file before
+ *                  deleting the old one.
+ */
+result FZ_PUBLIC_SYMBOL rename_file(native_string const& source, native_string const& dest, bool allow_copy = true);
 
 }
 
