@@ -1,5 +1,11 @@
 #include "libfilezilla/local_filesys.hpp"
-#ifndef FZ_WINDOWS
+
+#include "libfilezilla/buffer.hpp"
+#include "libfilezilla/file.hpp"
+
+#ifdef FZ_WINDOWS
+#include "windows/security_descriptor_builder.hpp"
+#else
 #include <errno.h>
 #include <sys/fcntl.h>
 #include <sys/stat.h>
@@ -210,13 +216,13 @@ local_filesys::type get_file_info_impl(int(*do_stat)(struct stat & buf, char con
 	return local_filesys::file;
 }
 
-local_filesys::type get_file_info_at(char const* path, DIR* dir, bool &is_link, int64_t* size, datetime* modification_time, int *mode)
+local_filesys::type get_file_info_at(char const* path, DIR* dir, bool &is_link, int64_t* size, datetime* modification_time, int *mode, bool follow)
 {
 	auto do_stat = [](struct stat & buf, char const* path, DIR * dir, bool follow)
 	{
 		return fstatat(dirfd(dir), path, &buf, follow ? 0 : AT_SYMLINK_NOFOLLOW);
 	};
-	return get_file_info_impl(do_stat, path, dir, is_link, size, modification_time, mode, true);
+	return get_file_info_impl(do_stat, path, dir, is_link, size, modification_time, mode, follow);
 }
 #endif
 
@@ -361,15 +367,17 @@ local_filesys::type local_filesys::get_file_info(native_string const& path, bool
 	return do_get_file_info(path, is_link, size, modification_time, mode, follow_links);
 }
 
-result local_filesys::begin_find_files(native_string path, bool dirs_only)
+result local_filesys::begin_find_files(native_string path, bool dirs_only, bool query_symlink_targets)
 {
+	end_find_files();
+
 	if (path.empty()) {
 		return result{result::nodir};
 	}
 
-	end_find_files();
-
 	dirs_only_ = dirs_only;
+	query_symlink_targets_ = query_symlink_targets;
+
 #ifdef FZ_WINDOWS
 	if (is_separator(path.back())) {
 		m_find_path = path;
@@ -385,7 +393,7 @@ result local_filesys::begin_find_files(native_string path, bool dirs_only)
 		has_next_ = false;
 		switch (GetLastError()) {
 			case ERROR_ACCESS_DENIED:
-				return result{result::noperm };
+				return result{result::noperm};
 			default:
 				return result{result::other};
 		}
@@ -416,6 +424,36 @@ result local_filesys::begin_find_files(native_string path, bool dirs_only)
 #endif
 }
 
+#if FZ_UNIX
+result local_filesys::begin_find_files(int fd, bool dirs_only, bool query_symlink_targets)
+{
+	end_find_files();
+
+	if (fd == -1) {
+		return {result::nodir};
+	}
+
+	dirs_only_ = dirs_only;
+	query_symlink_targets_ = query_symlink_targets;
+
+	dir_ = fdopendir(fd);
+	if (!dir_) {
+		switch (errno) {
+			case EACCES:
+			case EPERM:
+				return result{result::noperm};
+			case ENOTDIR:
+			case ENOENT:
+				return result{result::nodir};
+			default:
+				return result{result::other};
+		}
+	}
+
+	return result{result::ok};
+}
+#endif
+
 void local_filesys::end_find_files()
 {
 #ifdef FZ_WINDOWS
@@ -473,7 +511,7 @@ bool local_filesys::get_next_file(native_string& name)
 #if HAVE_STRUCT_DIRENT_D_TYPE
 			if (entry->d_type == DT_LNK) {
 				bool wasLink{};
-				if (get_file_info_at(entry->d_name, dir_, wasLink, nullptr, nullptr, nullptr) != dir) {
+				if (get_file_info_at(entry->d_name, dir_, wasLink, nullptr, nullptr, nullptr, query_symlink_targets_) != dir) {
 					continue;
 				}
 			}
@@ -483,7 +521,7 @@ bool local_filesys::get_next_file(native_string& name)
 #else
 			// Solaris doesn't have d_type
 			bool wasLink{};
-			if (get_file_info_at(entry->d_name, dir_, wasLink, nullptr, nullptr, nullptr) != dir) {
+			if (get_file_info_at(entry->d_name, dir_, wasLink, nullptr, nullptr, nullptr, query_symlink_targets_) != dir) {
 				continue;
 			}
 #endif
@@ -520,7 +558,7 @@ bool local_filesys::get_next_file(native_string& name, bool &is_link, local_file
 		t = (m_find_data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) ? dir : file;
 
 		is_link = (m_find_data.dwFileAttributes & FILE_ATTRIBUTE_REPARSE_POINT) != 0 && IsReparseTagNameSurrogate(m_find_data.dwReserved0);
-		if (is_link) {
+		if (is_link && query_symlink_targets_) {
 			// Follow the reparse point
 			HANDLE hFile = CreateFile((m_find_path + name).c_str(), FILE_READ_ATTRIBUTES | FILE_READ_EA, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, nullptr, OPEN_EXISTING, FILE_FLAG_BACKUP_SEMANTICS, nullptr);
 			if (hFile != INVALID_HANDLE_VALUE) {
@@ -610,7 +648,7 @@ bool local_filesys::get_next_file(native_string& name, bool &is_link, local_file
 #if HAVE_STRUCT_DIRENT_D_TYPE
 		if (dirs_only_) {
 			if (entry->d_type == DT_LNK) {
-				if (get_file_info_at(entry->d_name, dir_, is_link, size, modification_time, mode) != dir) {
+				if (get_file_info_at(entry->d_name, dir_, is_link, size, modification_time, mode, query_symlink_targets_) != dir) {
 					continue;
 				}
 
@@ -624,7 +662,7 @@ bool local_filesys::get_next_file(native_string& name, bool &is_link, local_file
 		}
 #endif
 
-		t = get_file_info_at(entry->d_name, dir_, is_link, size, modification_time, mode);
+		t = get_file_info_at(entry->d_name, dir_, is_link, size, modification_time, mode, query_symlink_targets_);
 		if (t == unknown) { // Happens for example in case of permission denied
 #if HAVE_STRUCT_DIRENT_D_TYPE
 			t = (entry->d_type == DT_DIR) ? dir : file;
@@ -734,75 +772,41 @@ native_string local_filesys::get_link_target(native_string const& path)
 }
 
 namespace {
-result do_mkdir(native_string const& path, bool current_user_only)
+result do_mkdir(native_string const& path, mkdir_permissions permissions)
 {
 	result ret{result::other};
 #ifdef FZ_WINDOWS
-	if (current_user_only) {
-		HANDLE token{ INVALID_HANDLE_VALUE };
-		BOOL res = OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, &token);
-		if (!res) {
-			return result{result::other};
+
+	SECURITY_ATTRIBUTES attr{};
+	attr.bInheritHandle = false;
+	attr.nLength = sizeof(SECURITY_ATTRIBUTES);
+
+	security_descriptor_builder sdb;
+	if (permissions != mkdir_permissions::normal) {
+		sdb.add(security_descriptor_builder::self);
+		if (permissions == mkdir_permissions::cur_user_and_admins) {
+			sdb.add(security_descriptor_builder::administrators);
 		}
-
-		DWORD needed{};
-		GetTokenInformation(token, TokenUser, NULL, 0, &needed);
-		if (GetLastError() != ERROR_INSUFFICIENT_BUFFER) {
-			return result{result::other};
+		auto sd = sdb.get_sd();
+		if (!sd) {
+			return {result::other};
 		}
-
-		PTOKEN_USER tu = static_cast<PTOKEN_USER>(malloc(needed));
-		if (tu) {
-			if (GetTokenInformation(token, TokenUser, tu, needed, &needed)) {
-				needed = sizeof(ACL) + ((sizeof(ACCESS_ALLOWED_ACE) - sizeof(DWORD)) + GetLengthSid(tu->User.Sid));
-				PACL acl = static_cast<PACL>(malloc(needed));
-				if (acl) {
-					if (InitializeAcl(acl, needed, ACL_REVISION)) {
-						if (AddAccessAllowedAceEx(acl, ACL_REVISION, CONTAINER_INHERIT_ACE | OBJECT_INHERIT_ACE, GENERIC_ALL | STANDARD_RIGHTS_ALL | SPECIFIC_RIGHTS_ALL, tu->User.Sid)) {
-							SECURITY_DESCRIPTOR sd;
-							InitializeSecurityDescriptor(&sd, SECURITY_DESCRIPTOR_REVISION);
-							//SetSecurityDescriptorControl(&sd, 
-							SetSecurityDescriptorDacl(&sd, TRUE, acl, FALSE);
-							SetSecurityDescriptorOwner(&sd, tu->User.Sid, FALSE);
-							SetSecurityDescriptorGroup(&sd, NULL, FALSE);
-							SetSecurityDescriptorSacl(&sd, FALSE, NULL, FALSE);
-
-							SECURITY_ATTRIBUTES attr;
-							attr.bInheritHandle = false;
-							attr.nLength = sizeof(SECURITY_ATTRIBUTES);
-							attr.lpSecurityDescriptor = &sd;
-
-							res = CreateDirectory(path.c_str(), &attr);
-							if (res) {
-								ret = result{result::ok};
-							}
-							else if (GetLastError() == ERROR_ACCESS_DENIED) {
-								ret = result{result::noperm};
-							}
-						}
-					}
-					free(acl);
-				}
-			}
-			free(tu);
-		}
-		CloseHandle(token);
+		attr.lpSecurityDescriptor = sd;
 	}
-	else {
-		if (CreateDirectory(path.c_str(), nullptr)) {
-			ret = result{result::ok};
-		}
-		else if (GetLastError() == ERROR_ACCESS_DENIED) {
-			ret = result{result::noperm};
-		}
+	
+	if (CreateDirectory(path.c_str(), &attr)) {
+		ret = {result::ok};
+	}
+	else if (GetLastError() == ERROR_ACCESS_DENIED) {
+		ret = {result::noperm};
 	}
 #else
-	int res = ::mkdir(path.c_str(), current_user_only ? 0700 : 0777);
+	int res = ::mkdir(path.c_str(), (permissions == mkdir_permissions::normal) ? 0777 : 0700);
 	if (!res) {
-		ret = result{result::ok};
+		ret = {result::ok};
 	}
 	else if (errno == EACCES || errno == EPERM) {
-		ret = result{result::noperm};
+		ret = {result::noperm};
 	}
 #endif
 
@@ -810,7 +814,7 @@ result do_mkdir(native_string const& path, bool current_user_only)
 }
 }
 
-result mkdir(native_string const& absolute_path, bool recurse, bool current_user_only, native_string* last_created)
+result mkdir(native_string const& absolute_path, bool recurse, mkdir_permissions permissions, native_string* last_created)
 {
 	// Step 0: Require an absolute path
 #ifdef FZ_WINDOWS
@@ -904,7 +908,7 @@ result mkdir(native_string const& absolute_path, bool recurse, bool current_user
 		for (size_t i = 0; i < segments.size(); ++i) {
 			work += local_filesys::path_separator;
 			work += segments[segments.size() - i - 1];
-			result r = do_mkdir(work, current_user_only && (i + 1) == segments.size());
+			result r = do_mkdir(work, ((i + 1) == segments.size()) ? permissions : mkdir_permissions::normal);
 			if (!r) {
 				return r;
 			}
@@ -914,7 +918,7 @@ result mkdir(native_string const& absolute_path, bool recurse, bool current_user
 		}
 	}
 	else {
-		result r = do_mkdir(absolute_path, current_user_only);
+		result r = do_mkdir(absolute_path, permissions);
 		if (!r) {
 			return r;
 		}
@@ -925,4 +929,139 @@ result mkdir(native_string const& absolute_path, bool recurse, bool current_user
 
 	return result{result::ok};
 }
+
+#ifndef FZ_WINDOWS
+namespace {
+static result do_copy(native_string const& source, native_string const& dest, bool & dest_opened)
+{
+	fz::file in(source, fz::file::reading, fz::file::existing);
+
+	if (!in.opened()) {
+		// TODO error codes
+		return {result::other};
+	}
+
+	fz::file out(dest, fz::file::writing, fz::file::empty);
+	if (!out.opened()) {
+		// TODO error codes
+		return {result::other};
+	}
+
+	dest_opened = true;
+
+	fz::buffer buffer;
+	while (true) {
+		if (buffer.empty()) {
+			auto read = in.read(buffer.get(64 * 1024), 64 * 1024);
+			if (read < 0) {
+				return {result::other};
+			}
+			else if (read) {
+				buffer.add(read);
+			}
+			else {
+				if (buffer.empty()) {
+					return {result::ok};
+				}
+			}
+		}
+		auto written = out.write(buffer.get(), buffer.size());
+		if (written <= 0) {
+			return {result::other};
+		}
+
+		buffer.consume(written);
+	}
+	if (!out.fsync()) {
+		return {result::other};
+	}
+
+	return {result::ok};
+}
+}
+#endif
+
+result rename_file(native_string const& source, native_string const& dest, bool allow_copy)
+{
+#ifdef FZ_WINDOWS
+	DWORD flags = MOVEFILE_REPLACE_EXISTING;
+	if (allow_copy) {
+		flags |= MOVEFILE_COPY_ALLOWED;
+	}
+	DWORD res = MoveFileExW(source.c_str(), dest.c_str(), flags);
+	if (res) {
+		return {result::ok};
+	}
+
+	DWORD err = GetLastError();
+	switch (err) {
+		case ERROR_FILE_NOT_FOUND:
+			return {result::nofile};
+		case ERROR_PATH_NOT_FOUND:
+		return {result::nodir};
+		case ERROR_ACCESS_DENIED:
+			return {result::noperm};
+		case ERROR_DISK_FULL:
+			return {result::nospace};
+		default:
+			return {result::other};
+	}
+#else
+	int res = rename(source.c_str(), dest.c_str());
+	if (!res) {
+		return {result::ok};
+	}
+
+	switch (errno) {
+	case EPERM:
+	case EACCES:
+		return {result::noperm};
+	case ENOSPC:
+		return {result::nospace};
+	case ENOTDIR:
+		return {result::nodir};
+	case ENOENT:
+	case EISDIR:
+		return {result::nofile};
+	case EXDEV:
+		break;
+	default:
+		return {result::other};
+	}
+
+	if (!allow_copy) {
+		return {result::other};
+	}
+
+	bool dest_opened{};
+	auto ret = do_copy(source, dest, dest_opened);
+	if (!ret) {
+		if (dest_opened) {
+			unlink(dest.c_str());
+		}
+		return ret;
+	}
+
+	res = unlink(source.c_str());
+	if (res != 0) {
+		switch (errno) {
+		case EPERM:
+		case EACCES:
+			return {result::noperm};
+		case ENOTDIR:
+			return {result::nodir};
+		case ENOENT:
+		case EISDIR:
+			return {result::nofile};
+		case EXDEV:
+			break;
+		default:
+			return {result::other};
+		}
+	}
+
+	return result{result::ok};
+#endif
+}
+
 }
