@@ -1,8 +1,9 @@
+#include "libfilezilla/impersonation.hpp"
 #include "libfilezilla/process.hpp"
 
 #ifdef FZ_WINDOWS
 
-#include "libfilezilla/private/windows.hpp"
+#include "libfilezilla/glue/windows.hpp"
 
 namespace fz {
 
@@ -134,6 +135,7 @@ native_string get_cmd_line(native_string const& cmd, std::vector<native_string>:
 }
 }
 
+HANDLE get_handle(impersonation_token const& t);
 class process::impl
 {
 public:
@@ -154,7 +156,7 @@ public:
 			err_.create(true);
 	}
 
-	bool spawn(native_string const& cmd, std::vector<native_string>::const_iterator const& begin, std::vector<native_string>::const_iterator const& end, bool redirect_io)
+	bool spawn(native_string const& cmd, std::vector<native_string>::const_iterator const& begin, std::vector<native_string>::const_iterator const& end, bool redirect_io, impersonation_token const* it = nullptr)
 	{
 		if (process_ != INVALID_HANDLE_VALUE) {
 			return false;
@@ -181,7 +183,13 @@ public:
 		auto cmdline_buf = cmdline.data();
 
 		DWORD const flags = CREATE_UNICODE_ENVIRONMENT | CREATE_DEFAULT_ERROR_MODE | CREATE_NO_WINDOW;
-		BOOL res = CreateProcess(cmd.c_str(), cmdline_buf, nullptr, nullptr, TRUE, flags, nullptr, nullptr, &si, &pi);
+		BOOL res;
+		if (it) {
+			 res = CreateProcessAsUser(get_handle(*it), cmd.c_str(), cmdline_buf, nullptr, nullptr, TRUE, flags, nullptr, nullptr, &si, &pi);
+		}
+		else {
+			res = CreateProcess(cmd.c_str(), cmdline_buf, nullptr, nullptr, TRUE, flags, nullptr, nullptr, &si, &pi);
+		}
 		if (!res) {
 			return false;
 		}
@@ -373,7 +381,7 @@ public:
 			err_.create();
 	}
 
-	bool spawn(native_string const& cmd, std::vector<native_string>::const_iterator const& begin, std::vector<native_string>::const_iterator const& end, bool redirect_io, std::vector<int> const& extra_fds = std::vector<int>())
+	bool spawn(native_string const& cmd, std::vector<native_string>::const_iterator const& begin, std::vector<native_string>::const_iterator const& end, bool redirect_io, std::vector<int> const& extra_fds = std::vector<int>(), impersonation_token const* it = nullptr)
 	{
 		if (pid_ != -1) {
 			return false;
@@ -421,6 +429,14 @@ public:
 				}
 			}
 
+#if FZ_UNIX
+			if (it && *it) {
+				if (!set_process_impersonation(*it)) {
+					_exit(1);
+				}
+			}
+#endif
+
 			// Execute process
 			execv(cmd.c_str(), argV.get()); // noreturn on success
 
@@ -514,11 +530,28 @@ bool process::spawn(native_string const& cmd, std::vector<native_string> const&
 	return impl_ ? impl_->spawn(cmd, args.cbegin(), args.cend(), redirect_io) : false;
 }
 
+#if FZ_WINDOWS || FZ_UNIX
+bool process::spawn(impersonation_token const& it, native_string const& cmd, std::vector<native_string> const& args, bool redirect_io)
+{
+#if FZ_WINDOWS
+	return impl_ ? impl_->spawn(cmd, args.cbegin(), args.cend(), redirect_io, &it) : false;
+#else
+	return impl_ ? impl_->spawn(cmd, args.cbegin(), args.cend(), redirect_io, {}, &it) : false;
+#endif
+}
+#endif
+
 #ifndef FZ_WINDOWS
 bool process::spawn(native_string const& cmd, std::vector<native_string> const& args, std::vector<int> const& extra_fds, bool redirect_io)
 {
 	return impl_ ? impl_->spawn(cmd, args.cbegin(), args.cend(), redirect_io, extra_fds) : false;
 }
+
+bool process::spawn(impersonation_token const& it, native_string const& cmd, std::vector<native_string> const& args, std::vector<int> const& extra_fds, bool redirect_io)
+{
+	return impl_ ? impl_->spawn(cmd, args.cbegin(), args.cend(), redirect_io, extra_fds, &it) : false;
+}
+
 #endif
 
 bool process::spawn(std::vector<native_string> const& command_with_args, bool redirect_io)
