@@ -1,7 +1,7 @@
 #include "libfilezilla/libfilezilla.hpp"
 
 #ifdef FZ_WINDOWS
-  #include "libfilezilla/private/windows.hpp"
+  #include "libfilezilla/glue/windows.hpp"
   #include <winsock2.h>
   #include <ws2tcpip.h>
   #include <mstcpip.h>
@@ -185,9 +185,8 @@ bool FZ_PRIVATE_SYMBOL has_pending_event(event_handler * handler, socket_event_s
 }
 #endif
 
-namespace {
 #ifdef FZ_WINDOWS
-static int convert_msw_error_code(int error)
+int convert_msw_error_code(int error)
 {
 	// Takes an MSW socket error and converts it into an equivalent POSIX error code.
 	switch (error)
@@ -236,7 +235,10 @@ static int convert_msw_error_code(int error)
 		return error;
 	}
 }
+#endif
 
+namespace {
+#ifdef FZ_WINDOWS
 int last_socket_error()
 {
 	return convert_msw_error_code(WSAGetLastError());
@@ -570,6 +572,10 @@ public:
 protected:
 	static socket::socket_t create_socket_fd(addrinfo const& addr)
 	{
+#ifndef FZ_WINDOWS
+		disable_sigpipe();
+#endif
+
 		socket::socket_t fd;
 #if defined(SOCK_CLOEXEC) && !defined(FZ_WINDOWS)
 		fd = ::socket(addr.ai_family, addr.ai_socktype | SOCK_CLOEXEC, addr.ai_protocol);
@@ -1509,6 +1515,11 @@ int listen_socket::listen(address_type family, int port)
 				continue;
 			}
 
+			if (addr->ai_family == AF_INET6) {
+				int enable = 1;
+				setsockopt(fd_, IPPROTO_IPV6, IPV6_V6ONLY, reinterpret_cast<char const*>(&enable), sizeof(enable));
+			}
+
 			res = ::bind(fd_, addr->ai_addr, addr->ai_addrlen);
 			if (!res) {
 				break;
@@ -1582,6 +1593,10 @@ socket_descriptor listen_socket::fast_accept(int &error)
 #endif
 		}
 
+		if (fd == -1) {
+			error = errno;
+		}
+
 #if !defined(FZ_WINDOWS) && !defined(HAVE_POLL)
 		if (fd >= FD_SETSIZE) {
 			::close(fd);
@@ -1756,15 +1771,6 @@ int socket::write(void const* buffer, unsigned int size, int& error)
 	const int flags = MSG_NOSIGNAL;
 #else
 	const int flags = 0;
-
-#if !defined(SO_NOSIGPIPE) && !defined(FZ_WINDOWS)
-	// Some systems have neither. Need to block signal
-	struct sigaction old_action;
-	struct sigaction action = {};
-	action.sa_handler = SIG_IGN;
-	int signal_set = sigaction(SIGPIPE, &action, &old_action);
-#endif
-
 #endif
 
 #if DEBUG_SOCKETEVENTS
@@ -1777,13 +1783,6 @@ int socket::write(void const* buffer, unsigned int size, int& error)
 
 	int res = send(fd_, (const char*)buffer, size, flags);
 
-#if !defined(MSG_NOSIGNAL) && !defined(SO_NOSIGPIPE) && !defined(FZ_WINDOWS)
-	// Restore previous signal handler
-	if (!signal_set) {
-		sigaction(SIGPIPE, &old_action, 0);
-	}
-#endif
-
 	if (res == -1) {
 		error = last_socket_error();
 		if (error == EAGAIN) {
@@ -2030,4 +2029,14 @@ int socket_layer::shutdown_read()
 {
 	return next_layer_.shutdown_read();
 }
+
+socket_base::socket_t socket::get_descriptor()
+{
+	if (!socket_thread_) {
+		return fd_;
+	}
+
+	scoped_lock l(socket_thread_->mutex_);
+	return fd_; // Mutex as fd_ might change during connect
+}
 }
