@@ -26,9 +26,9 @@ bool tls_layer::client_handshake(event_handler* const verification_handler, std:
 	return impl_->client_handshake(session_to_resume, session_hostname, std::vector<uint8_t>(), verification_handler);
 }
 
-bool tls_layer::server_handshake(std::vector<uint8_t> const& session_to_resume, std::string_view const& preamble)
+bool tls_layer::server_handshake(std::vector<uint8_t> const& session_to_resume, std::string_view const& preamble, tls_server_flags flags)
 {
-	return impl_->server_handshake(session_to_resume, preamble);
+	return impl_->server_handshake(session_to_resume, preamble, flags);
 }
 
 int tls_layer::read(void *buffer, unsigned int size, int& error)
@@ -96,7 +96,7 @@ bool tls_layer::set_certificate_file(native_string const& keyfile, native_string
 	return impl_->set_certificate_file(keyfile, certsfile, password, pem);
 }
 
-bool tls_layer::set_certificate(std::string const& key, std::string const& certs, native_string const& password, bool pem)
+bool tls_layer::set_certificate(std::string_view const& key, std::string_view const& certs, native_string const& password, bool pem)
 {
 	return impl_->set_certificate(key, certs, password, pem);
 }
@@ -131,6 +131,11 @@ std::pair<std::string, std::string> tls_layer::generate_selfsigned_certificate(n
 	return tls_layer_impl::generate_selfsigned_certificate(password, distinguished_name, hostnames);
 }
 
+std::pair<std::string, std::string> tls_layer::generate_csr(native_string const& password, std::string const& distinguished_name, std::vector<std::string> const& hostnames, bool csr_as_pem)
+{
+	return tls_layer_impl::generate_csr(password, distinguished_name, hostnames, csr_as_pem);
+}
+
 int tls_layer::shutdown_read()
 {
 	return impl_->shutdown_read();
@@ -185,4 +190,22 @@ bool tls_layer::is_server() const
 	return impl_ ? impl_->server_ : false;
 }
 
+void tls_layer::set_min_tls_ver(tls_ver ver)
+{
+	if (impl_) {
+		impl_->set_min_tls_ver(ver);
+	}
+}
+
+void tls_layer::set_max_tls_ver(tls_ver ver)
+{
+	if (impl_) {
+		impl_->set_max_tls_ver(ver);
+	}
+}
+
+int tls_layer::new_session_ticket()
+{
+	return impl_ ? impl_->new_session_ticket() : false;
+}
 }
