@@ -6,18 +6,36 @@ typedef std::make_signed_t<size_t> ssize_t;
 #endif
 
 #include <gnutls/gnutls.h>
+#include <gnutls/x509.h>
 
 #include "libfilezilla/buffer.hpp"
 #include "libfilezilla/logger.hpp"
 #include "libfilezilla/socket.hpp"
 #include "libfilezilla/tls_info.hpp"
+#include "libfilezilla/tls_layer.hpp"
+
+#include <optional>
 
 namespace fz {
 class tls_system_trust_store;
 class logger_interface;
-namespace {
-struct cert_list_holder;
-}
+
+struct cert_list_holder final
+{
+	cert_list_holder() = default;
+	~cert_list_holder() {
+		for (unsigned int i = 0; i < certs_size; ++i) {
+			gnutls_x509_crt_deinit(certs[i]);
+		}
+		gnutls_free(certs);
+	}
+
+	cert_list_holder(cert_list_holder const&) = delete;
+	cert_list_holder& operator=(cert_list_holder const&) = delete;
+
+	gnutls_x509_crt_t * certs{};
+	unsigned int certs_size{};
+};
 
 class tls_layer;
 class tls_layer_impl final
@@ -28,7 +46,7 @@ public:
 
 	bool client_handshake(std::vector<uint8_t> const& session_to_resume, native_string const& session_hostname, std::vector<uint8_t> const& required_certificate, event_handler * verification_handler);
 
-	bool server_handshake(std::vector<uint8_t> const& session_to_resume, std::string_view const& preamble);
+	bool server_handshake(std::vector<uint8_t> const& session_to_resume, std::string_view const& preamble, tls_server_flags flags);
 
 	int connect(native_string const& host, unsigned int port, address_type family);
 
@@ -58,7 +76,7 @@ public:
 
 	bool set_certificate_file(native_string const& keyfile, native_string const& certsfile, native_string const& password, bool pem);
 
-	bool set_certificate(std::string const& key, std::string const& certs, native_string const& password, bool pem);
+	bool set_certificate(std::string_view const& key, std::string_view const& certs, native_string const& password, bool pem);
 
 	static std::string get_gnutls_version();
 
@@ -66,6 +84,7 @@ public:
 	ssize_t pull_function(void* data, size_t len);
 
 	static std::pair<std::string, std::string> generate_selfsigned_certificate(native_string const& password, std::string const& distinguished_name, std::vector<std::string> const& hostnames);
+	static std::pair<std::string, std::string> generate_csr(native_string const& password, std::string const& distinguished_name, std::vector<std::string> const& hostnames, bool csr_as_pem);
 
 	int shutdown_read();
 
@@ -74,11 +93,17 @@ public:
 	std::string get_alpn() const;
 	native_string get_hostname() const;
 
+	static int load_certificates(std::string_view const& in, bool pem, gnutls_x509_crt_t *& certs, unsigned int & certs_size, bool & sort);
+	static bool extract_cert(gnutls_x509_crt_t const& cert, x509_certificate& out, bool last, logger_interface * logger);
+
+	void set_min_tls_ver(tls_ver ver);
+	void set_max_tls_ver(tls_ver ver);
+
 private:
 	bool init();
 	void deinit();
 
-	bool init_session(bool client);
+	bool init_session(bool client, int extra_flags = 0);
 	void deinit_session();
 
 	int continue_write();
@@ -106,19 +131,16 @@ private:
 
 	bool get_sorted_peer_certificates(gnutls_x509_crt_t *& certs, unsigned int & certs_size);
 
-	bool extract_cert(gnutls_x509_crt_t const& cert, x509_certificate& out, bool last);
-	std::vector<x509_certificate::subject_name> get_cert_subject_alt_names(gnutls_x509_crt_t cert);
+	static std::vector<x509_certificate::subject_name> get_cert_subject_alt_names(gnutls_x509_crt_t cert);
 
 	void log_verification_error(int status);
 
 	void set_hostname(native_string const& host);
 
-	bool is_client() const {
-		return ticket_key_.empty();
-	}
-
 	bool do_set_alpn();
 
+	int new_session_ticket();
+
 	tls_layer& tls_layer_;
 
 	logger_interface & logger_;
@@ -126,6 +148,8 @@ private:
 	gnutls_session_t session_{};
 
 	std::vector<uint8_t> ticket_key_;
+	std::vector<uint8_t> session_db_key_;
+	std::vector<uint8_t> session_db_data_;
 
 	gnutls_certificate_credentials_t cert_credentials_{};
 
@@ -162,6 +186,9 @@ private:
 
 	event_handler * verification_handler_{};
 
+	tls_ver min_tls_ver_{tls_ver::v1_0};
+	std::optional<tls_ver> max_tls_ver_;
+
 	int socket_error_{}; // Set in the push and pull functions if reading/writing fails fatally
 	bool socket_eof_{};
 
@@ -170,11 +197,16 @@ private:
 
 	bool write_blocked_by_send_buffer_{};
 
+	bool send_new_ticket_{};
+
 #if DEBUG_SOCKETEVENTS
 	bool debug_can_read_{};
 	bool debug_can_write_{};
 #endif
 };
+
+std::string read_certificates_file(native_string const& certsfile, logger_interface * logger);
+
 }
 
 #endif
